# https://github.com/Klipper3d/klipper/blob/master/klippy/extras/thermistor.py

class Thermistor:
    def __init__(self, pullup: float, inline_resistor: float) -> None: ...
    def setup_coefficients_beta(self, t1: float, r1: float, beta: float) -> None: ...
    def calc_temp(self, adc: float) -> float: ...
